"""This module implements the MachineModel class."""
from __future__ import annotations

from bqskit.ir.location import CircuitLocation
from bqskit.qis.graph import CouplingGraph
from bqskit.qis.graph import CouplingGraphLike
from bqskit.utils.typing import is_integer


class MachineModel:
    """A model of a quantum processing unit's connectivity."""

    def __init__(
        self,
        num_qudits: int,
        coupling_graph: CouplingGraphLike | None = None,
    ) -> None:
        """
        MachineModel Constructor.

        Args:
            num_qudits (int): The total number of qudits in the machine.

            coupling_graph (Iterable[tuple[int, int]] | None): A coupling
                graph describing which pairs of qudits can interact.
                Given as an undirected edge set. If left as None, then
                an all-to-all coupling graph is used as a default.
                (Default: None)

        Raises:
            ValueError: If `num_qudits` is nonpositive.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                f'Expected integer num_qudits, got {type(num_qudits)}.',
            )

        if num_qudits <= 0:
            raise ValueError(f'Expected positive num_qudits, got {num_qudits}.')

        if coupling_graph is None:
            coupling_graph = CouplingGraph.all_to_all(num_qudits)

        if not CouplingGraph.is_valid_coupling_graph(
                coupling_graph, num_qudits,
        ):
            raise TypeError('Invalid coupling graph.')

        self.coupling_graph = CouplingGraph(coupling_graph)
        self.num_qudits = num_qudits

    def get_locations(self, block_size: int) -> list[CircuitLocation]:
        """Return all `block_size` connected blocks of qudit indicies."""
        return self.coupling_graph.get_subgraphs_of_size(block_size)



"""This module implements the ZXZXZDecomposition."""
from __future__ import annotations

import cmath
from typing import Any

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import RZGate
from bqskit.ir.gates import SqrtXGate


class ZXZXZDecomposition(BasePass):
    """
    The ZXZXZDecomposition class.

    Convert a single-qubit circuit to ZXZXZ sequence.
    """

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if circuit.num_qudits != 1:
            raise ValueError(
                'Cannot convert multi-qudit circuit into ZXZXZ sequence.',
            )

        if circuit.radixes[0] != 2:
            raise ValueError(
                'Cannot convert non-qubit circuit into ZXZXZ sequence.',
            )

        utry = circuit.get_unitary()

        # Calculate params
        utry = np.linalg.det(utry) ** (-0.5) * utry
        i1 = cmath.phase(utry[1, 1])
        i2 = cmath.phase(utry[1, 0])
        t = 2 * np.arctan2(abs(utry[1, 0]), abs(utry[0, 0])) + np.pi
        p = i1 + i2 + np.pi
        l = i1 - i2

        # Move angles into [-pi, pi)
        t = (t + np.pi) % (2 * np.pi) - np.pi
        p = (p + np.pi) % (2 * np.pi) - np.pi
        l = (l + np.pi) % (2 * np.pi) - np.pi

        new_circuit = Circuit(1)
        new_circuit.append_gate(RZGate(), 0, [l])
        new_circuit.append_gate(SqrtXGate(), 0)
        new_circuit.append_gate(RZGate(), 0, [t])
        new_circuit.append_gate(SqrtXGate(), 0)
        new_circuit.append_gate(RZGate(), 0, [p])
        circuit.become(new_circuit)



"""This module implements the PassAlias class."""
from __future__ import annotations
import abc

from typing import Any, Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit


class PassAlias(BasePass):
    """A pass that is an alias for another pass or sequence of passes."""
    
    @abc.abstractmethod
    def get_passes(self) -> list[BasePass]:
        """Return the passes that should be run, when this alias is called."""

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        for bqskit_pass in self.get_passes():
            bqskit_pass.run(circuit, data)



class PassGroup(PassAlias):
    """A pass that is a group of other passes."""

    def __init__(self, passes: BasePass | Sequence[BasePass]) -> None:
        self.passes = passes
    
    def get_passes(self) -> list[BasePass]:
        return self.passes


"""This module implements the PermutationMatrix class."""
from __future__ import annotations

import logging
from functools import lru_cache
from typing import Sequence

import numpy as np
from bqskitrs import calc_permutation_matrix

from bqskit.ir.location import CircuitLocation
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class PermutationMatrix(UnitaryMatrix):
    """A binary, unitary matrix with a single 1 in each row and column."""

    def __init__(
        self,
        input: UnitaryLike,
        radixes: Sequence[int] = [],
        check_arguments: bool = True,
    ) -> None:
        """
        Constructs a `PermutationMatrix` from the supplied matrix.

        See :class:`UnitaryMatrix` for more info.
        """
        if check_arguments and not PermutationMatrix.is_permutation(input):
            raise ValueError('Input failed permutation condition.')

        super().__init__(input, radixes, check_arguments)

    @staticmethod
    @lru_cache(maxsize=None)
    def from_qubit_location(
        num_qubits: int,
        location: tuple[int, ...],
    ) -> PermutationMatrix:
        """
        Creates the permutation matrix specified by arguments.

        The resulting matrix will move the first len(location) qubits
        into positions defined by location.

        Args:
            num_qubits (int): Total number of qubits

            location (tuple[int, ...]): The desired locations to swap
                the starting qubits to.

        Returns:
            PermutationMatrix: A 2**num_qubits by 2**num_qubits permutation
            matrix.

        Examples:
            >>> calc_permutation_matrix( 2, (0, 1) )
                [ [ 1, 0, 0, 0 ],
                  [ 0, 1, 0, 0 ],
                  [ 0, 0, 1, 0 ],
                  [ 0, 0, 0, 1 ] ]

            Here the 4x4 identity is returned because there are 2 total
            qubits, specified by the first parameter, and the desired
            permutation is [0, 1] -> [0, 1].

            >>> calc_permutation_matrix( 2, (1,) ) = # Also equals
            >>> calc_permutation_matrix( 2, (1, 0) ) =
                [ [ 1, 0, 0, 0 ],
                  [ 0, 0, 1, 0 ],
                  [ 0, 1, 0, 0 ],
                  [ 0, 0, 0, 1 ] ]

            This is a more interesting example. The swap gate's unitary
            is returned here since we are working with 2 qubits and want
            the permutation that swaps the both qubits. This is given by
            the permutation [0] -> [1] and [0, 1] -> [1, 0] in the second
            case. Both calls produce identical permutations.
        """

        if not isinstance(num_qubits, int):
            raise TypeError(
                'Expected integer num_qudits'
                ', got %s.' % type(num_qubits),
            )

        if num_qubits <= 0:
            raise ValueError(
                'Expected positive num_qudits'
                ', got %d.' % num_qubits,
            )

        if not CircuitLocation.is_location(location, num_qubits):
            raise TypeError('Invalid location.')

        matrix = calc_permutation_matrix(num_qubits, location)
        return PermutationMatrix(matrix)

    @staticmethod
    def is_permutation(P: np.typing.ArrayLike, tol: float = 1e-8) -> bool:
        """
        Check if P is a permutation matrix.

        Args:
            P (np.typing.ArrayLike): The matrix to check.

            tol (float): The numerical precision of the check.

        Returns:
            bool: True if P is a permutation matrix.
        """

        if isinstance(P, PermutationMatrix):
            return True

        if not UnitaryMatrix.is_unitary(P, tol):
            return False

        if not isinstance(P, np.ndarray):
            P = np.array(P)

        if not all(s == 1 for s in P.sum(0)):
            _logger.debug('Not all rows sum to 1.')
            return False

        if not all(s == 1 for s in P.sum(1)):
            _logger.debug('Not all columns sum to 1.')
            return False

        if not all(e == 1 or e == 0 for row in P for e in row):
            _logger.debug('Not all elements are 0 or 1.')
            return False

        return True



"""This package implements synthesis passes and synthesis related classes."""
from __future__ import annotations

from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.qfast import QFASTDecompositionPass
from bqskit.passes.synthesis.qpredict import QPredictDecompositionPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass

__all__ = [
    'LEAPSynthesisPass',
    'QFASTDecompositionPass',
    'QPredictDecompositionPass',
    'QSearchSynthesisPass',
    'SynthesisPass',
]



"""This module implements the Compiler class."""
from __future__ import annotations

import logging
import uuid
from typing import Any

from dask.distributed import Client
from dask.distributed import Future

from bqskit.compiler.executor import Executor
from bqskit.compiler.task import CompilationTask
from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class Compiler:
    """
    The BQSKit compiler class.

    A compiler is responsible for accepting and managing compilation tasks.
    The compiler class spins up a Dask execution environment, which
    compilation tasks can then access to parallelize their operations.
    The compiler is implemented as a context manager and it is recommended
    to use it as one. If the compiler is not used in a context manager, it
    is the responsibility of the user to call `close()`.

    Examples:
        >>> with Compiler() as compiler:
        ...     circuit = compiler.compile(task)

        >>> compiler = Compiler()
        >>> circuit = compiler.compile(task)
        >>> compiler.close()
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Construct a Compiler object.

        Notes:
            All arguments are passed directly to Dask. You can use
            these to connect to and configure a Dask cluster.
        """
        dask_options = {
            'silence_logs': logging.getLogger('bqskit').level,
        }
        if 'address' not in kwargs and 'scheduler_file' not in kwargs:
            dask_options['threads_per_worker'] = 1

        dask_options.update(kwargs)

        self.client = Client(*args, **dask_options)
        self.tasks: dict[uuid.UUID, Future] = {}
        _logger.info('Started compiler process.')

    def __enter__(self) -> Compiler:
        """Enter a context for this compiler."""
        return self

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        """Shutdown compiler."""
        self.close()

    def close(self) -> None:
        """Shutdown compiler."""
        try:
            self.client.close()
            self.tasks = {}
            _logger.info('Stopped compiler process.')
        except (AttributeError, TypeError):
            pass

    def submit(self, task: CompilationTask) -> None:
        """Submit a CompilationTask to the Compiler."""
        executor = self.client.scatter(Executor(task))
        future = self.client.submit(Executor.run, executor, pure=False)
        self.tasks[task.task_id] = future
        _logger.info('Submitted task: %s' % task.task_id)

    def status(self, task: CompilationTask) -> str:
        """Retrieve the status of the specified CompilationTask."""
        return self.tasks[task.task_id].status

    def result(self, task: CompilationTask) -> Circuit:
        """Block until the CompilationTask is finished, return its result."""
        circ = self.tasks[task.task_id].result()[0]
        return circ

    def cancel(self, task: CompilationTask) -> None:
        """Remove a task from the compiler's workqueue."""
        self.client.cancel(self.tasks[task.task_id])
        _logger.info('Cancelled task: %s' % task.task_id)

    def compile(self, task: CompilationTask) -> Circuit:
        """Submit and execute the CompilationTask, block until its done."""
        _logger.info('Compiling task: %s' % task.task_id)
        self.submit(task)
        result = self.result(task)
        return result

    def analyze(self, task: CompilationTask, key: str) -> Any:
        """Gather the value associated with `key` in the task's data."""
        if task.task_id not in self.tasks:
            self.submit(task)
        return self.tasks[task.task_id].result()[1][key]



"""This module implements the Minimization class."""
from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.ir.opt.instantiater import Instantiater
from bqskit.ir.opt.minimizer import Minimizer
from bqskit.ir.opt.minimizers.ceres import CeresMinimizer
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit


class Minimization(Instantiater):
    """The Minimization circuit instantiater."""

    def __init__(
        self,
        cost_fn_gen: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        minimizer: Minimizer | None = None,
        **kwargs: dict[str, Any],  # TODO: handle dist_tol and other options
    ) -> None:
        """
        Construct and configure a Minimization Instantiater.

        Args:
            cost_fn_gen (CostFunctionGenerator): This generator creates cost
                functions that are minimized.
                (Default: HilbertSchmidtGenerator())

            minimizer (Minimizer | None): The minimizer to use. If left as
                None, attempts to select best one.
        """

        if not isinstance(cost_fn_gen, CostFunctionGenerator):
            raise TypeError(
                'Expected CostFunctionGenerator, got %s.' % type(cost_fn_gen),
            )

        if minimizer is not None and not isinstance(minimizer, Minimizer):
            raise TypeError('Expected Minimizer, got %s.' % type(minimizer))

        # Default to the fast CeresMinimizer
        if minimizer is None:
            minimizer = CeresMinimizer()

        self.cost_fn_gen = cost_fn_gen
        self.minimizer = minimizer

    def instantiate(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector,
        x0: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """Instantiate `circuit`, see Instantiater for more info."""
        cost = self.cost_fn_gen.gen_cost(circuit, target)
        return self.minimizer.minimize(cost, x0)

    @staticmethod
    def is_capable(circuit: Circuit) -> bool:
        """Return true if the circuit can be instantiated."""
        return True

    @staticmethod
    def get_violation_report(circuit: Circuit) -> str:
        """
        All circuits can be instantiated with minimization.

        See Instantiater for more info.
        """
        raise ValueError('Circuit can be instantiated.')

    @staticmethod
    def get_method_name() -> str:
        """Return the name of this method."""
        return 'minimization'



"""This module implements the CNOTToCZPass."""
from __future__ import annotations

from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import CircuitGate
from bqskit.ir.gates import CNOTGate
from bqskit.ir.gates import CZGate
from bqskit.ir.gates import HGate
from bqskit.ir.operation import Operation


class CNOTToCZPass(BasePass):
    """
    The CNOTToCZPass class.

    This uses a rule to convert CNOTs to CZs.
    """

    def __init__(self) -> None:
        """Construct a CNOTToCZPass."""
        circuit = Circuit(2)
        circuit.append_gate(HGate(), 1)
        circuit.append_gate(CZGate(), (0, 1))
        circuit.append_gate(HGate(), 1)
        self.cg = CircuitGate(circuit)

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Find all cnots
        cnot_points = []
        for cycle, op in circuit.operations_with_cycles():
            if isinstance(op.gate, CNOTGate):
                cnot_points.append((cycle, op.location[0]))

        # Create new operations
        ops = [
            Operation(
                self.cg,
                circuit[p].location,
                self.cg._circuit.params,
            )
            for p in cnot_points
        ]

        # Replace cnots with new ops
        circuit.batch_replace(cnot_points, ops)
        circuit.unfold_all()  # TODO: Replace with batch_unfold



"""This module implements the U8Gate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qutritgate import QutritGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class U8Gate(QutritGate, DifferentiableUnitary, CachedClass):
    """The U8 single qutrit gate."""

    _num_qudits = 1
    _num_params = 8

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])
        p2 = np.exp(1j * params[4])
        m2 = np.exp(-1j * params[4])
        p3 = np.exp(1j * params[5])
        m3 = np.exp(-1j * params[5])
        p4 = np.exp(1j * params[6])
        m4 = np.exp(-1j * params[6])
        p5 = np.exp(1j * params[7])
        m5 = np.exp(-1j * params[7])

        return UnitaryMatrix(
            [
                [
                    c1 * c2 * p1,
                    s1 * p3,
                    c1 * s2 * p4,
                ],
                [
                    s2 * s3 * m4 * m5 - s1 * c2 * c3 * p1 * p2 * m3,
                    c1 * c3 * p2,
                    -c2 * s3 * m1 * m5 - s1 * s2 * c3 * p2 * m3 * p4,
                ],
                [
                    -s1 * c2 * s3 * p1 * m3 * p5 - s2 * c3 * m2 * m4,
                    c1 * s3 * p5,
                    c2 * c3 * m1 * m2 - s1 * s2 * s3 * m3 * p4 * p5,
                ],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        s1 = np.sin(params[0])
        c1 = np.cos(params[0])
        s2 = np.sin(params[1])
        c2 = np.cos(params[1])
        s3 = np.sin(params[2])
        c3 = np.cos(params[2])

        p1 = np.exp(1j * params[3])
        m1 = np.exp(-1j * params[3])
        p2 = np.exp(1j * params[4])
        m2 = np.exp(-1j * params[4])
        p3 = np.exp(1j * params[5])
        m3 = np.exp(-1j * params[5])
        p4 = np.exp(1j * params[6])
        m4 = np.exp(-1j * params[6])
        p5 = np.exp(1j * params[7])
        m5 = np.exp(-1j * params[7])

        return np.array(
            [
                [  # wrt params[0]
                    [-s1 * c2 * p1, c1 * p3, -s1 * s2 * p4],
                    [
                        -c1 * c2 * c3 * p1 * p2 * m3, -s1 * \
                        c3 * p2, -c1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        -c1 * c2 * s3 * p1 * m3 * p5, -s1 * \
                        s3 * p5, -c1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[1]
                    [-c1 * s2 * p1, 0, c1 * c2 * p4],
                    [
                        c2 * s3 * m4 * m5 + s1 * s2 * c3 * p1 * p2 * m3, 0,
                        s2 * s3 * m1 * m5 - s1 * c2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        s1 * s2 * s3 * p1 * m3 * p5 - c2 * c3 * m2 * m4, 0,
                        -s2 * c3 * m1 * m2 - s1 * c2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[2]
                    [0, 0, 0],
                    [
                        s2 * c3 * m4 * m5 + s1 * c2 * s3 * p1 * p2 * m3,
                        -c1 * s3 * p2,
                        -c2 * c3 * m1 * m5 + s1 * s2 * s3 * p2 * m3 * p4,
                    ],
                    [
                        -s1 * c2 * c3 * p1 * m3 * p5 + s2 * s3 * m2 * m4,
                        c1 * c3 * p5,
                        -c2 * s3 * m1 * m2 - s1 * s2 * c3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[3]
                    [1j * c1 * c2 * p1, 0, 0],
                    [
                        -1j * s1 * c2 * c3 * p1 * p2 * m3,
                        0,
                        1j * c2 * s3 * m1 * m5,
                    ],
                    [
                        -1j * s1 * c2 * s3 * p1 * m3 * p5,
                        0, -1j * c2 * c3 * m1 * m2,
                    ],
                ],

                [  # wrt params[4]
                    [0, 0, 0],
                    [
                        -1j * s1 * c2 * c3 * p1 * p2 * m3, 1j * c1 * \
                        c3 * p2, -1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [1j * s2 * c3 * m2 * m4, 0, -1j * c2 * c3 * m1 * m2],
                ],

                [  # wrt params[5]
                    [0, 1j * s1 * p3, 0],
                    [
                        1j * s1 * c2 * c3 * p1 * p2 * m3, 0,
                        1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        1j * s1 * c2 * s3 * p1 * m3 * p5, 0,
                        1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[6]
                    [0, 0, 1j * c1 * s2 * p4],
                    [
                        -1j * s2 * s3 * m4 * m5,
                        0,
                        -1j * s1 * s2 * c3 * p2 * m3 * p4,
                    ],
                    [
                        1j * s2 * c3 * m2 * m4,
                        0,
                        -1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],

                [  # wrt params[7]
                    [0, 0, 0],
                    [-1j * s2 * s3 * m4 * m5, 0, 1j * c2 * s3 * m1 * m5],
                    [
                        -1j * s1 * c2 * s3 * p1 * m3 * p5, 1j * c1 * \
                        s3 * p5, -1j * s1 * s2 * s3 * m3 * p4 * p5,
                    ],
                ],
            ],
        )



"""This module implements the FourParamGenerator class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates import CNOTGate
from bqskit.ir.gates import RXGate
from bqskit.ir.gates import RYGate
from bqskit.ir.gates import RZGate
from bqskit.ir.gates import U3Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class FourParamGenerator(LayerGenerator):
    """
    The FourParamGenerator class.

    This is an optimized layer generator that uses commutativity rules
    to reduce the number of parameters per block. This also fixes the
    gate set to use cnots, ry, rz, and u3 gates. This is based on the
    following equivalences:

    U--C--U   U--C--Rz--Ry--Rz   U--C--Ry--Rz
       |    ~    |             ~    |
    U--X--U   U--X--Rx--Ry--Rx   U--X--Ry--Rx
    """

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector,
        data: dict[str, Any],
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` is not qubit only.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if not target.is_qubit_only():
            raise ValueError('Cannot generate layers for non-qubit circuits.')

        init_circuit = Circuit(target.num_qudits, target.radixes)
        for i in range(init_circuit.num_qudits):
            init_circuit.append_gate(U3Gate(), [i])
        return init_circuit

    def gen_successors(
        self,
        circuit: Circuit,
        data: dict[str, Any],
    ) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the machine model
        model = BasePass.get_model(circuit, data)

        # Generate successors
        successors = []
        for edge in model.coupling_graph:
            successor = circuit.copy()
            successor.append_gate(CNOTGate(), [edge[0], edge[1]])
            successor.append_gate(RYGate(), edge[0])
            successor.append_gate(RZGate(), edge[0])
            successor.append_gate(RYGate(), edge[1])
            successor.append_gate(RXGate(), edge[1])
            successors.append(successor)

        return successors



"""This module implements the ParallelDo class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable
from typing import Sequence

from distributed import as_completed
from distributed import worker_client

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.utils.typing import is_sequence

_logger = logging.getLogger(__name__)


class ParallelDo(BasePass):
    """
    The ParallelDo class.

    This is a control pass that executes a group of pass sequences in parallel.
    """

    def __init__(
        self,
        pass_sequences: Sequence[BasePass | Sequence[BasePass]],
        less_than: Callable[[Circuit, Circuit], bool],
        pick_fisrt: bool = False,
    ) -> None:
        """
        Construct a ParallelDo.

        Args:
            pass_sequences (Sequence[BasePass | Sequence[BasePass]]):
                The group of pass sequences to run in parallel.

            less_than (Callable[[Circuit, Circuit], bool]): Return True
                if the first circuit is preferred to the second one.
                This will be used to determine which output circuit to
                select.

            pick_first (bool): If true, then the pass will complete as
                soon as one of the subsequences finishes and will return
                the first result.

        Raises:
            ValueError: If a Sequence[BasePass] is given, but it is empty.
        """
        if not is_sequence(pass_sequences):
            raise TypeError(
                'Expected sequence of sequences of passes, got %s.'
                % type(pass_sequences),
            )

        self.pass_seqs = []
        for ps in pass_sequences:
            if not is_sequence(ps) and not isinstance(ps, BasePass):
                raise TypeError(
                    'Expected Pass or sequence of Passes, got %s.'
                    % type(ps),
                )

            if is_sequence(ps):
                truth_list = [isinstance(elem, BasePass) for elem in ps]
                if not all(truth_list):
                    raise TypeError(
                        'Expected Pass or sequence of Passes, got %s.'
                        % type(ps[truth_list.index(False)]),
                    )
                if len(ps) == 0:
                    raise ValueError('Expected at least one pass.')

            self.pass_seqs.append(ps if is_sequence(ps) else [ps])

        self.less_than = less_than
        self.pick_first = pick_fisrt

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug('Running pass sequences in parallel.')

        if 'parallel' not in data:
            raise RuntimeError(
                'Parallel pass cannot be executed without a compiler.',
            )

        with worker_client() as client:
            futures = []
            for pass_seq in self.pass_seqs:
                future = client.submit(_sub_do_work, pass_seq, circuit, data)
                futures.append(future)

            if self.pick_first:
                _, result = next(as_completed(futures, with_results=True))
                for future in futures:
                    if not future.done:
                        client.cancel(future)

                circuit.become(result[0])
                data.update(result[1])

            else:
                best_circ = None
                best_data = None
                for future in futures:
                    _circ, _data = future.result()
                    if best_circ is None or self.less_than(_circ, best_circ):
                        best_circ = _circ
                        best_data = _data
                if best_circ is None or best_data is None:
                    raise RuntimeError('No valid circuit found.')
                circuit.become(best_circ)
                data.update(best_data)


def _sub_do_work(
    loop_body: Sequence[BasePass],
    circuit: Circuit,
    data: dict[str, Any],
) -> tuple[Circuit, dict[str, Any]]:
    for loop_pass in loop_body:
        loop_pass.run(circuit, data)
    return circuit, data



"""This module implements the QSearchSynthesisPass."""
from __future__ import annotations

import logging
from typing import Any

from dask.distributed import get_client
from dask.distributed import rejoin
from dask.distributed import secede

from bqskit.ir.circuit import Circuit
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.passes.search.frontier import Frontier
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators import SimpleLayerGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.passes.search.heuristics import AStarHeuristic
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.unitary import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number


_logger = logging.getLogger(__name__)


class QSearchSynthesisPass(SynthesisPass):
    """
    A pass implementing the QSearch A* synthesis algorithm.

    References:
        Davis, Marc G., et al. “Towards Optimal Topology Aware Quantum
        Circuit Synthesis.” 2020 IEEE International Conference on Quantum
        Computing and Engineering (QCE). IEEE, 2020.
    """

    def __init__(
        self,
        heuristic_function: HeuristicFunction = AStarHeuristic(),
        layer_generator: LayerGenerator = SimpleLayerGenerator(),
        success_threshold: float = 1e-10,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        max_layer: int | None = None,
        store_partial_solutions: bool = False,
        partials_per_depth: int = 25,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Construct a search-based synthesis pass.

        Args:
            heuristic_function (HeuristicFunction): The heuristic to guide
                search.

            layer_generator (LayerGenerator): The successor function
                to guide node expansion.

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the cost function. (Default: 1e-10)

            cost (CostFunction | None): The cost function that determines
                distance during synthesis. The goal of this synthesis pass
                is to implement circuits for the given unitaries that have
                a cost less than the `success_threshold`.
                (Default: HSDistance())

            max_layer (int): The maximum number of layers to append without
                success before termination. If left as None it will default
                to unlimited. (Default: None)

            store_partial_solutions (bool): Whether to store partial solutions
                at different depths inside of the data dict. (Default: False)

            partials_per_depth (int): The maximum number of partials
                to store per search depth. No effect if
                `store_partial_solutions` is False. (Default: 25)

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

        Raises:
            ValueError: If `max_depth` is nonpositive.
        """
        if not isinstance(heuristic_function, HeuristicFunction):
            raise TypeError(
                'Expected HeursiticFunction, got %s.'
                % type(heuristic_function),
            )

        if not isinstance(layer_generator, LayerGenerator):
            raise TypeError(
                'Expected LayerGenerator, got %s.'
                % type(layer_generator),
            )

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if max_layer is not None and not is_integer(max_layer):
            raise TypeError(
                'Expected max_layer to be an integer, got %s' % type(max_layer),
            )

        if max_layer is not None and max_layer <= 0:
            raise ValueError(
                'Expected max_layer to be positive, got %d.' % int(max_layer),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.heuristic_function = heuristic_function
        self.layer_gen = layer_generator
        self.success_threshold = success_threshold
        self.cost = cost
        self.max_layer = max_layer
        self.instantiate_options: dict[str, Any] = {'cost_fn_gen': self.cost}
        self.instantiate_options.update(instantiate_options)
        self.store_partial_solutions = store_partial_solutions
        self.partials_per_depth = partials_per_depth

    def synthesize(self, utry: UnitaryMatrix, data: dict[str, Any]) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        frontier = Frontier(utry, self.heuristic_function)

        # Seed the search with an initial layer
        initial_layer = self.layer_gen.gen_initial_layer(utry, data)
        initial_layer.instantiate(utry, **self.instantiate_options)
        frontier.add(initial_layer, 0)

        # Track best circuit, initially the initial layer
        search_data: dict[str, Any] = {}
        search_data['best_dist'] = self.cost.calc_cost(initial_layer, utry)
        search_data['best_circ'] = initial_layer
        search_data['best_layer'] = 0
        if self.store_partial_solutions:
            search_data['psols'] = {}
        _logger.info(
            'Search started, initial layer has cost: %e.' %
            search_data['best_dist'],
        )

        if search_data['best_dist'] < self.success_threshold:
            _logger.info('Successful synthesis.')
            return initial_layer

        if 'parallel' in data:  # In Parallel
            client = get_client()
            while not frontier.empty():
                top_circuit, layer = frontier.pop()

                # Generate successors and evaluate each
                successors = self.layer_gen.gen_successors(top_circuit, data)

                # Submit instantiate jobs
                futures = client.map(
                    Circuit.instantiate,
                    successors,
                    target=utry,
                    parallel=True,
                    **self.instantiate_options,
                    pure=False,
                )

                # Wait for and gather results
                secede()
                circuits = client.gather(futures)
                rejoin()

                for circuit in circuits:
                    if self.evaluate_node(
                        circuit,
                        utry,
                        data,
                        frontier,
                        layer,
                        search_data,
                    ):
                        if self.store_partial_solutions:
                            data['psols'] = search_data['psols']
                        return circuit

        else:  # Sequentially
            while not frontier.empty():
                top_circuit, layer = frontier.pop()

                # Generate successors and evaluate each
                successors = self.layer_gen.gen_successors(top_circuit, data)

                for circuit in successors:
                    circuit.instantiate(utry, **self.instantiate_options)
                    if self.evaluate_node(
                        circuit,
                        utry,
                        data,
                        frontier,
                        layer,
                        search_data,
                    ):
                        if self.store_partial_solutions:
                            data['psols'] = search_data['psols']
                        return circuit

        _logger.info('Frontier emptied.')
        _logger.info(
            'Returning best known circuit with %d layer%s and cost: %e.'
            % (
                search_data['best_layer'], '' if search_data['best_layer'] == 1
                else 's', search_data['best_dist'],
            ),
        )
        if self.store_partial_solutions:
            data['psols'] = search_data['psols']

        return search_data['best_circ']

    def evaluate_node(
        self,
        circuit: Circuit,
        utry: UnitaryMatrix,
        data: dict[str, Any],
        frontier: Frontier,
        layer: int,
        search_data: dict[str, Any],
    ) -> bool:
        dist = self.cost.calc_cost(circuit, utry)

        if dist < self.success_threshold:
            _logger.info('Successful synthesis.')
            return True

        if dist < search_data['best_dist']:
            _logger.info(
                'New best circuit found with %d layer%s and cost: %e.'
                % (layer + 1, '' if layer == 0 else 's', dist),
            )
            search_data['best_dist'] = dist
            search_data['best_circ'] = circuit
            search_data['best_layer'] = layer

        if self.store_partial_solutions:
            if layer not in search_data['psols']:
                search_data['psols'][layer] = []

            search_data['psols'][layer].append((circuit.copy(), dist))

            if len(search_data['psols'][layer]) > self.partials_per_depth:
                search_data['psols'][layer].sort(key=lambda x: x[1])
                del search_data['psols'][layer][-1]

        if self.max_layer is None or layer + 1 < self.max_layer:
            frontier.add(circuit, layer + 1)

        return False



"""This package contains LayerGenerator definitions."""
from __future__ import annotations

from bqskit.passes.search.generators.fourparam import FourParamGenerator
from bqskit.passes.search.generators.middleout import MiddleOutLayerGenerator
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.passes.search.generators.single import SingleQuditLayerGenerator
from bqskit.passes.search.generators.stair import StairLayerGenerator

__all__ = [
    'SimpleLayerGenerator',
    'SeedLayerGenerator',
    'StairLayerGenerator',
    'MiddleOutLayerGenerator',
    'FourParamGenerator',
    'SingleQuditLayerGenerator',
]



"""This package implements BQSKit's rule-based passes."""
from __future__ import annotations

from bqskit.passes.rules.cnot2cz import CNOTToCZPass
from bqskit.passes.rules.zxzxz import ZXZXZDecomposition

__all__ = ['CNOTToCZPass', 'ZXZXZDecomposition']



"""This module implements the SeedLayerGenerator class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.ir.circuit import Circuit
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
_logger = logging.getLogger(__name__)


class SeedLayerGenerator(LayerGenerator):
    """Layer Generator for search that starts from a seed."""

    def __init__(
        self,
        seed: Circuit,
        forward_generator: LayerGenerator = SimpleLayerGenerator(),
        num_removed: int = 1,
    ) -> None:
        """
        Construct a SeedLayerGenerator.

        Args:
            seed (Circuit): The seed to start from.

            forward_generator (Gate): A generator used to grow the circuit.

            num_removed (int): The number of gates removed from the circuit
                in each backwards branch.
        """
        if not isinstance(seed, Circuit):
            raise TypeError(f'Expected Circuit for seed, got {type(seed)}')

        if not isinstance(forward_generator, LayerGenerator):
            raise TypeError(
                'Expected LayerGenerator for forward_generator'
                f', got {type(forward_generator)}.',
            )

        if not is_integer(num_removed):
            raise TypeError(
                f'Expected integer for num_removed, got {type(num_removed)}.',
            )

        self.seed = seed
        self.forward_generator = forward_generator
        self.num_removed = num_removed

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector,
        data: dict[str, Any],
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a size or radix mismatch with
                `self.seed`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if target.dim != self.seed.dim:
            raise ValueError('Seed dimension mismatch with target.')

        data['seed_seen_before'] = {self.hash_structure(self.seed)}

        return self.seed

    def gen_successors(
        self,
        circuit: Circuit,
        data: dict[str, Any],
    ) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        if circuit.dim != self.seed.dim:
            raise ValueError('Seed dimension mismatch with circuit.')

        # Generate successors
        successors = self.forward_generator.gen_successors(circuit, data)

        removed_count = 0
        for cycle, op in circuit.operations_with_cycles(reverse=True):
            removed_count += 1
            if removed_count > self.num_removed:
                break
            copied_circuit = circuit.copy()
            copied_circuit.pop((cycle, op.location[0]))
            successors.insert(0, copied_circuit)

        filtered_successors = []
        for s in successors:
            h = self.hash_structure(s)
            if h not in data['seed_seen_before']:
                data['seed_seen_before'].add(h)
                filtered_successors.append(s)

        return filtered_successors

    @staticmethod
    def hash_structure(circuit: Circuit) -> int:
        hashes = []
        for cycle, op in circuit.operations_with_cycles():
            hashes.append(hash((cycle, str(op))))
            if len(hashes) > 100:
                hashes = [sum(hashes)]
        return sum(hashes)



"""This module implements the SingleQuditLayerGenerator class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
_logger = logging.getLogger(__name__)


class SingleQuditLayerGenerator(LayerGenerator):
    """
    The SingleQuditLayerGenerator class.

    This module implements a layer generator for a single-qudit circuits with
    hetergenous gate support.
    """

    def __init__(self, gates: list[Gate], allow_repeats: bool = False) -> None:
        """
        Construct a SingleQuditLayerGenerator.

        Args:
            gates (Gate): The single-qudit gate set.

            allow_repeats (bool): By default this is set to False,
                and the generator will not generate a circuit that
                has the same gate twice in a row. Some gates are not
                closed under multiplication and it may be beneficial
                to allow them to repeat. In this case, `allow_repeats`
                should be True.

        Raises:
            ValueError: If any pair of gates from `gates` has a
                different radix.

            ValueError: If any gate from `gates` is not single-qudit.

            ValueError: If `gates` is empty.
        """
        self.radix = None
        for gate in gates:
            if not isinstance(gate, Gate):
                raise TypeError(f'Expected gate, got {type(gate)}.')

            if gate.num_qudits > 1:
                raise ValueError(f'Expected single-qudit gate, got {gate}.')

            if self.radix is None:
                self.radix = gate.radixes[0]

            elif gate.radixes[0] != self.radix:
                raise ValueError('Gate set has a radix inconsistency.')

        if len(gates) == 0:
            raise ValueError('Empty gate set.')

        self.gates = gates
        self.allow_repeats = allow_repeats

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector,
        data: dict[str, Any],
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a radix mismatch with
                `self.gates`.

            ValueError: If `target` is not single-qudit.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if target.num_qudits > 1:
            raise ValueError('Target is larger than a single-qudit.')

        if target.radixes[0] != self.radix:
            raise ValueError('Mismatch between generator and target radix.')

        init_circuit = Circuit(target.num_qudits, target.radixes)
        return init_circuit

    def gen_successors(
        self,
        circuit: Circuit,
        data: dict[str, Any],
    ) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is not a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits > 1:
            raise ValueError('Cannot expand a multi-qudit circuit.')

        # Record last gate
        last_gate = circuit[-1, -1].gate if len(circuit) >= 1 else None

        # Generate successors
        successors = []
        for gate in self.gates:

            # Skip adding a gate that repeats the last one on the circuit
            if not self.allow_repeats and gate == last_gate:
                continue

            successor = circuit.copy()
            successor.append_gate(gate, 0)
            successors.append(successor)

        return successors



"""This module implements the CouplingGraph class."""
from __future__ import annotations

import itertools as it
import logging
from functools import lru_cache
from typing import Any
from typing import Collection
from typing import Iterable
from typing import Iterator
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_iterable

_logger = logging.getLogger(__name__)


class CouplingGraph(Collection[Tuple[int, int]]):
    """A graph representing connections in a qudit topology."""

    def __init__(self, graph: Iterable[tuple[int, int]]) -> None:
        if isinstance(graph, CouplingGraph):
            self.num_qudits: int = graph.num_qudits
            self._edges: set[tuple[int, int]] = graph._edges
            self._adj: list[list[int]] = graph._adj
            return

        if not CouplingGraph.is_valid_coupling_graph(graph):
            raise TypeError('Invalid coupling graph.')

        self._edges = set(graph)

        self.num_qudits = 0
        for q1, q2 in self._edges:
            self.num_qudits = max(self.num_qudits, max(q1, q2))
        self.num_qudits += 1

        self._adj = [[] for _ in range(self.num_qudits)]
        for q1, q2 in self._edges:
            self._adj[q1].append(q2)
            self._adj[q2].append(q1)

    def is_fully_connected(self) -> bool:
        """Return true if the graph is fully connected."""
        frontier = {0}
        qudits_seen = {0}

        while len(frontier) > 0:
            expanded_qudits = set()

            for qudit in frontier:
                expanded_qudits.update(self._adj[qudit])

            frontier = expanded_qudits - qudits_seen
            qudits_seen.update(expanded_qudits)

            if len(qudits_seen) == self.num_qudits:
                return True

        return False

    def get_neighbors_of(self, qudit: int) -> list[int]:
        """Return the qudits adjacent to `qudit`."""
        return self._adj[qudit]

    def __contains__(self, __o: object) -> bool:
        return self._edges.__contains__(__o)

    def __eq__(self, __o: object) -> bool:
        return self._edges.__eq__(__o)

    def __iter__(self) -> Iterator[tuple[int, int]]:
        return self._edges.__iter__()

    def __hash__(self) -> int:
        return hash(tuple(self._edges))

    def __len__(self) -> int:
        return self._edges.__len__()

    def __str__(self) -> str:
        return 'CouplingGraph(' + self._edges.__str__() + ')'

    def __repr__(self) -> str:
        return self._edges.__repr__()

    def get_qudit_degrees(self) -> list[int]:
        return [len(l) for l in self._adj]

    def get_subgraph(
        self,
        location: CircuitLocationLike,
        renumbering: dict[int, int] | None = None,
    ) -> CouplingGraph:
        """Returns the sub-coupling-graph with qudits in `location`."""
        if not CircuitLocation.is_location(location, self.num_qudits):
            raise TypeError('Invalid location.')

        location = CircuitLocation(location)
        if renumbering is None:
            renumbering = {x: x for x in range(self.num_qudits)}

        subgraph = []
        for q0, q1 in self._edges:
            if q0 in location and q1 in location:
                subgraph.append((renumbering[q0], renumbering[q1]))
        return CouplingGraph(subgraph)

    @lru_cache(maxsize=None)
    def get_subgraphs_of_size(self, size: int) -> list[CircuitLocation]:
        """
        Find all sets of indices that form connected subgraphs on their own.

        Each location describes a valid spot for a `size`-sized gate,
        so the number of qudit_indices in each location is `size`.
        A location is only included if each pair of qudits is directly
        connected or connected through other qudits in the location.

        Args:
            size (int): The size of each location in the final list.

        Returns:
            list[CircuitLocation]: The locations compliant with the machine.

        Raises:
            ValueError: If `size` is nonpositive or too large.

        Notes:
            Does a breadth first search on all pairs of qudits, keeps paths
            that have length equal to size. Note that the coupling map
            is assumed to be undirected.
        """
        if not is_integer(size):
            raise TypeError(
                f'Expected integer for block_size, got {type(size)}',
            )

        if size > self.num_qudits:
            raise ValueError(
                'The block_size is too large; '
                f'expected <= {self.num_qudits}, got {size}.',
            )

        if size <= 0:
            raise ValueError(f'Nonpositive size; got {size}.')

        locations: set[CircuitLocation] = set()

        for qudit in range(self.num_qudits):
            # Get every valid set containing `qudit` with size == size
            self._location_search(locations, set(), qudit, size)

        return list(locations)

    def _location_search(
        self,
        locations: set[CircuitLocation],
        path: set[int],
        vertex: int,
        limit: int,
    ) -> None:
        """
        Add paths with length equal to limit to the `locations` set.

        Args:
            locations (set[CircuitLocation]): A list that contains all paths
                found so far of length equal to `limit`.

            path (set[int]): The qudit vertices currently included in
                the path.

            vertex (int): The vertex in the graph currently being examined.

            limit (int): The desired length of paths in the `locations`
                list.
        """
        if vertex in path:
            return

        curr_path = path.copy()
        curr_path.add(vertex)

        if len(curr_path) == limit:
            locations.add(CircuitLocation(list(curr_path)))
            return

        frontier: set[int] = {
            qudit
            for node in curr_path
            for qudit in self._adj[node]
            if qudit not in curr_path
        }

        for neighbor in frontier:
            self._location_search(locations, curr_path, neighbor, limit)

    @staticmethod
    def is_valid_coupling_graph(
        coupling_graph: Any,
        num_qudits: int | None = None,
    ) -> TypeGuard[CouplingGraphLike]:
        """
        Return true if the coupling graph is valid.

        Args:
            coupling_graph (Any): The coupling graph to check.

            num_qudits (int | None): The total number of qudits. All qudits
                should be less than this. If None, don't check.

        Returns:
            bool: Valid or not
        """

        if not is_iterable(coupling_graph):
            _logger.debug('Coupling graph is not iterable.')
            return False

        if len(list(coupling_graph)) == 0:
            return True

        if not all(isinstance(pair, tuple) for pair in coupling_graph):
            _logger.debug('Coupling graph is not a sequence of tuples.')
            return False

        if not all([len(pair) == 2 for pair in coupling_graph]):
            _logger.debug('Coupling graph is not a sequence of pairs.')
            return False

        if num_qudits is not None:
            if not (is_integer(num_qudits) and num_qudits > 0):
                _logger.debug('Invalid num_qudits in coupling graph check.')
                return False

            if not all(
                qudit < num_qudits
                for pair in coupling_graph
                for qudit in pair
            ):
                _logger.debug('Coupling graph has invalid qudits.')
                return False

        if not all([
            len(pair) == len(set(pair))
            for pair in coupling_graph
        ]):
            _logger.debug('Coupling graph has an invalid pair.')
            return False

        return True

    @staticmethod
    def all_to_all(num_qudits: int) -> CouplingGraph:
        return CouplingGraph(set(it.combinations(range(num_qudits), 2)))

    @staticmethod
    def linear(num_qudits: int) -> CouplingGraph:
        return CouplingGraph([(x, x + 1) for x in range(num_qudits - 1)])


CouplingGraphLike = Union[Iterable[Tuple[int, int]], CouplingGraph]



from __future__ import annotations

import pytest

from bqskit.compiler.machine import MachineModel


class TestMachineConstructor:
    def test_coupling_graph(self) -> None:
        coupling_graph = {(0, 1), (1, 2), (2, 3)}
        t = MachineModel(4, coupling_graph)

        assert len(t.coupling_graph) == 3
        for link in coupling_graph:
            assert link in t.coupling_graph

    def test_num_qudits(self) -> None:
        for n in [1, 2, 3, 4]:
            t = MachineModel(n)
            assert t.num_qudits == n

    def test_alltoall_2(self) -> None:
        t = MachineModel(2)
        assert len(t.coupling_graph) == 1
        assert (0, 1) in t.coupling_graph

    def test_alltoall_3(self) -> None:
        t = MachineModel(3)
        assert len(t.coupling_graph) == 3
        assert (0, 1) in t.coupling_graph
        assert (0, 2) in t.coupling_graph
        assert (1, 2) in t.coupling_graph

    def test_alltoall_4(self) -> None:
        t = MachineModel(4)
        assert len(t.coupling_graph) == 6
        assert (0, 1) in t.coupling_graph
        assert (0, 2) in t.coupling_graph
        assert (0, 3) in t.coupling_graph
        assert (1, 2) in t.coupling_graph
        assert (1, 3) in t.coupling_graph
        assert (2, 3) in t.coupling_graph

    def test_num_qudits_invalid(self) -> None:
        with pytest.raises(TypeError):
            MachineModel('a')  # type: ignore
        with pytest.raises(ValueError):
            MachineModel(0)

    def test_coupling_graph_invalid(self) -> None:
        coupling_graph = {(0, 1), (1, 2), (2, 3)}
        with pytest.raises(TypeError):
            MachineModel(2, coupling_graph)
        with pytest.raises(TypeError):
            MachineModel(2, (0, 1))  # type: ignore
        with pytest.raises(TypeError):
            MachineModel(2, 0)  # type: ignore
        with pytest.raises(TypeError):
            MachineModel(2, 'a')  # type: ignore



"""This package contains parameterized gates."""
from __future__ import annotations

from bqskit.ir.gates.parameterized.crx import CRXGate
from bqskit.ir.gates.parameterized.cry import CRYGate
from bqskit.ir.gates.parameterized.crz import CRZGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rxx import RXXGate
from bqskit.ir.gates.parameterized.ry import RYGate
from bqskit.ir.gates.parameterized.ryy import RYYGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.rzz import RZZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.ir.gates.parameterized.u2 import U2Gate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.gates.parameterized.u8 import U8Gate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate

__all__ = [
    'CRXGate',
    'CRYGate',
    'CRZGate',
    'PauliGate',
    'RXGate',
    'RXXGate',
    'RYGate',
    'RYYGate',
    'RZGate',
    'RZZGate',
    'U1Gate',
    'U2Gate',
    'U3Gate',
    'U8Gate',
    'VariableUnitaryGate',
]



"""This module implements the ControlledGate class."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class ControlledGate(
    ComposedGate,
    QubitGate,
    DifferentiableUnitary,
):
    """
    An arbitrary controlled gate.

    Given any qubit gate, ControlledGate can add control qubits.
    """

    def __init__(
        self,
        gate: Gate,
        num_controls: int = 1,
    ) -> None:
        """
        Construct a ControlledGate.

        Args:
            gate (Gate): The gate to control.

            num_controls (int): The number of controls to add.

        Raises:
            ValueError: If `num_controls` is less than 1
        """

        if not isinstance(gate, Gate):
            raise TypeError('Expected gate object, got %s.' % type(gate))

        if not is_integer(num_controls):
            raise TypeError(
                'Expected integer for num_controls, got %s.'
                % type(num_controls),
            )

        if num_controls < 1:
            raise ValueError(
                'Expected positive integer for num_controls, got %d.'
                % num_controls,
            )

        self.gate = gate
        self._num_qudits = gate.num_qudits + num_controls
        self.num_controls = num_controls
        self._name = '%d-Controlled(%s)' % (num_controls, gate.name)
        self._num_params = gate.num_params

        self.Ic = np.identity(2 ** num_controls, dtype=np.complex128)
        self.It = np.identity(gate.dim, dtype=np.complex128)
        self.OneProj = np.zeros(self.Ic.shape, dtype=np.complex128)
        self.OneProj[-1, -1] = 1
        self.left = np.kron((self.Ic - self.OneProj), self.It)

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0:
            U = self.gate.get_unitary()
            right = np.kron(self.OneProj, U)
            self.utry = UnitaryMatrix(self.left + right, self.radixes)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, 'utry'):
            return self.utry

        U = self.gate.get_unitary(params)
        right = np.kron(self.OneProj, U)
        return UnitaryMatrix(self.left + right, self.radixes)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return np.array([])

        grads = self.gate.get_grad(params)  # type: ignore
        right = np.kron(self.OneProj, grads)
        return self.left + right

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return self.utry, np.array([])

        U, G = self.gate.get_unitary_and_grad(params)  # type: ignore

        right = np.kron(self.OneProj, U)
        utry = UnitaryMatrix(self.left + right, self.radixes)
        grads = np.kron(self.OneProj, G)
        return utry, grads

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ControlledGate)
            and self.gate == other.gate
            and self.num_controls == other.num_controls
        )

    def __hash__(self) -> int:
        return hash((self.gate, self.num_controls))



"""
================================================
Compiler Infrastructure (:mod:`bqskit.compiler`)
================================================

The `bqskit.compiler` package implements the BQSKit's compiler framework.
This includes classes for defining, combining, executing, and managing
compilation algorithms.

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:

    BasePass
    Compiler
    CompilationTask
    MachineModel
"""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.compiler import Compiler
from bqskit.compiler.executor import Executor
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.task import CompilationTask

__all__ = [
    'BasePass',
    'Compiler',
    'Executor',
    'CompilationTask',
    'MachineModel',
]



"""This module implements the StairLayerGenerator class."""
from __future__ import annotations

import logging
from typing import Any

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


_logger = logging.getLogger(__name__)


class StairLayerGenerator(LayerGenerator):
    """Layer Generator for search that builds circuits from a single gate."""

    def __init__(self, gate: Gate) -> None:
        """
        Construct a StairLayerGenerator.

        Args:
            gate (Gate): The gate to build from.
        """
        if not isinstance(gate, Gate):
            raise TypeError(f'Expected Gate for gate, got {type(gate)}')

        self.gate = gate

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector,
        data: dict[str, Any],
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a size or radix mismatch with
                `self.seed`.
        """

        if not isinstance(target, (UnitaryMatrix, StateVector)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        init_circuit = Circuit(target.num_qudits, target.radixes)

        # Place RXX Gates on consective pairs of qudits
        nq = self.gate.num_qudits
        for i in range(init_circuit.num_qudits - (nq - 1)):
            init_circuit.append_gate(self.gate, [i + j for j in range(nq)])

        return init_circuit

    def gen_successors(
        self,
        circuit: Circuit,
        data: dict[str, Any],
    ) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        nq = self.gate.num_qudits
        successor = circuit.copy()
        for i in range(circuit.num_qudits - (nq - 1)):
            successor.append_gate(self.gate, [i + j for j in range(nq)])

        return [successor]



"""This module implements the U2Gate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class U2Gate(QubitGate, DifferentiableUnitary, CachedClass):
    """
    The U2 single qubit gate.

    It is given by the following parameterized unitary:

    .. math::

        \\begin{pmatrix}
        \\frac{\\sqrt{2}}{2} & -\\exp({i\\theta_1})\\frac{\\sqrt{2}}{2} \\\\
        \\exp({i\\theta_0})\\frac{\\sqrt{2}}{2} &
         \\exp({i(\\theta_0 + \\theta_1)})\\frac{\\sqrt{2}}{2} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _num_params = 2
    _qasm_name = 'u2'

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        sq2 = np.sqrt(2) / 2
        eip = np.exp(1j * params[0])
        eil = np.exp(1j * params[1])

        return UnitaryMatrix(
            [
                [sq2, -eil * sq2],
                [eip * sq2, eip * eil * sq2],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        sq2 = np.sqrt(2) / 2
        eip = np.exp(1j * params[0])
        eil = np.exp(1j * params[1])
        deip = 1j * np.exp(1j * params[0])
        deil = 1j * np.exp(1j * params[1])

        return np.array(
            [
                [  # wrt params[0]
                    [0, 0],
                    [deip * sq2, deip * eil * sq2],
                ],

                [  # wrt params[1]
                    [0, -deil * sq2],
                    [0, eip * deil * sq2],
                ],
            ], dtype=np.complex128,
        )



"""This module implements the BasePass abstract base class."""
from __future__ import annotations

import abc
import logging
from typing import Any

from bqskit.compiler.machine import MachineModel
from bqskit.ir.circuit import Circuit
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

_logger = logging.getLogger(__name__)


class BasePass(abc.ABC):
    """
    The abstract base for BQSKit compiler passes.

    All BQSKit algorithms must inherit from BasePass to be run within
    the compiler framework. Each child class will need to implement its
    algorithm inside of the :func:`run` method.

    Examples:
        >>> class PrintCNOTCountPass(BasePass):
        ...     def run(self, circ: Circuit, data: dict[str, Any] = {}) -> None:
        ...         print(f"Number of CNOTs: {circ.count(CNOTGate())}")
    """

    @property
    def name(self) -> str:
        """The name of the pass."""
        return self.__class__.__name__

    @abc.abstractmethod
    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """
        Perform the pass's operation on `circuit`.

        Args:
            circuit (Circuit): The circuit to operate on.

            data (Dict[str, Any]): Associated data for the pass.
                Can be used to get auxillary information from previous
                passes and to store information for future passes.
                This function should never error based on what is in
                this dictionary.

        Note:
            - This function should be self-contained and have no side effects.
              This is because it will be potentially run in parallel.
        """

    @staticmethod
    def get_model(circuit: Circuit, data: dict[str, Any]) -> MachineModel:
        """
        Retrieve the machine model from the data dictionary.

        Args:
            circuit (Circuit): The pass circuit.

            data (dict[str, Any]): The data dictionary.

        Returns:
            MachineModel: The machine model in the data dictionary, or
                a default one.
        """

        if len(data) == 0:
            return MachineModel(circuit.num_qudits)

        if 'machine_model' not in data:
            data['machine_model'] = MachineModel(circuit.num_qudits)

        if (
            not isinstance(data['machine_model'], MachineModel)
            or data['machine_model'].num_qudits < circuit.num_qudits
        ):
            _logger.warning('Expected machine_model to be a valid model.')
            return MachineModel(circuit.num_qudits)

        return data['machine_model']

    @staticmethod
    def get_target(circuit: Circuit, data: dict[str, Any]) -> UnitaryMatrix:
        """
        Retrieve the target unitary from the data dictionary.

        Args:
            circuit (Circuit): The pass circuit.

            data (dict[str, Any]): The data dictionary.

        Returns:
            UnitaryMatrix: The target unitary.
        """
        if len(data) == 0:
            return circuit.get_unitary()

        if 'target_unitary' not in data:
            data['target_unitary'] = circuit.get_unitary()

        if not isinstance(data['target_unitary'], UnitaryMatrix):
            _logger.warning('Expected target_unitary to be a unitary.')
            return circuit.get_unitary()

        return data['target_unitary']



"""This module implements the SimpleLayoutPass class."""
from __future__ import annotations

from typing import Any

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.ir.circuit import Circuit


class SimpleLayoutPass(BasePass):
    """
    Assigns logical qudits to physical qudits.

    Layout also encodes a machine model into the compilation workflow.
    """

    def __init__(self, model: MachineModel | None = None) -> None:
        """
        Construct a SimpleLayoutPass.

        Args:
            model (MachineModel | None): The machine model to use for
                layout and to encode into the compilation workflow.
                Defaults to an all-to-all model matching the size of the
                circuit.
        """
        if model is not None and not isinstance(model, MachineModel):
            raise TypeError(f'Expected MachineModel, got: {type(model)}.')

        self.model = model

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if self.model is None:
            data['machine_model'] = MachineModel(circuit.num_qudits)
            return

        if self.model.num_qudits < circuit.num_qudits:
            raise RuntimeError('Machine model is too small for circuit.')

        qudits = list(range(circuit.num_qudits))
        graph = self.model.coupling_graph.get_subgraph(qudits)
        if not graph.is_fully_connected():
            raise RuntimeError('Simple layout creates a disconnected graph.')

        data['machine_model'] = MachineModel(circuit.num_qudits, graph)



"""
====================================================================
Compilation and Synthesis Algorithms (:mod:`bqskit.passes`)
====================================================================

.. rubric:: Partitioning Passes

These passes will group together gates in the circuit. It is recommended
to use the QuickPartitioner currently.

.. autosummary::
    :toctree: autogen
    :recursive:

    ClusteringPartitioner
    GreedyPartitioner
    ScanPartitioner
    QuickPartitioner
    GroupSingleQuditGatePass

.. rubric:: Synthesis Passes

These passes perform synthesis. Decomposition passes break down
unitaries into smaller ones and will need to be followed up by another
synthesis pass to convert the circuit to native gates.

.. autosummary::
    :toctree: autogen
    :recursive:

    LEAPSynthesisPass
    QSearchSynthesisPass
    QFASTDecompositionPass
    QPredictDecompositionPass

.. rubric:: Processing Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    ScanningGateRemovalPass
    WindowOptimizationPass
    SubstitutePass

.. rubric:: Control Passes

This are passes that are used to manage control flow inside of a
compilation task.

.. autosummary::
    :toctree: autogen
    :recursive:

    DoWhileLoopPass
    ForEachBlockPass
    IfThenElsePass
    WhileLoopPass
    DoThenDecide
    ParallelDo

.. rubric:: Predicates

This objects are designed as conditions for use with control passes.

.. autosummary::
    :toctree: autogen
    :recursive:

    PassPredicate
    ChangePredicate
    GateCountPredicate
    NotPredicate

.. rubric:: Rule-based Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    CNOTToCZPass
    ZXZXZDecomposition

.. rubric:: Layout Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    SimpleLayoutPass

.. rubric:: Utility Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    CompressPass
    RecordStatsPass
    SetRandomSeedPass
    UnfoldPass
    UpdateDataPass
    ToU3Pass
    BlockConversionPass

.. rubric:: IO Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    LoadCheckpointPass
    SaveCheckpointPass
    SaveIntermediatePass
    RestoreIntermediatePass

.. rubric:: Search Heuristics

.. autosummary::
    :toctree: autogen
    :recursive:

    HeuristicFunction
    AStarHeuristic
    GreedyHeuristic
    DijkstraHeuristic

.. rubric:: Search Layer Generators

.. autosummary::
    :toctree: autogen
    :recursive:

    LayerGenerator
    SimpleLayerGenerator
    SeedLayerGenerator
    StairLayerGenerator
    SingleQuditLayerGenerator
    MiddleOutLayerGenerator
"""
from __future__ import annotations

from bqskit.passes.control.dothendecide import DoThenDecide
from bqskit.passes.control.dowhileloop import DoWhileLoopPass
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.control.ifthenelse import IfThenElsePass
from bqskit.passes.control.paralleldo import ParallelDo
from bqskit.passes.control.predicate import PassPredicate
from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate
from bqskit.passes.control.whileloop import WhileLoopPass
from bqskit.passes.io.checkpoint import LoadCheckpointPass
from bqskit.passes.io.checkpoint import SaveCheckpointPass
from bqskit.passes.io.intermediate import RestoreIntermediatePass
from bqskit.passes.io.intermediate import SaveIntermediatePass
from bqskit.passes.layout.simple import SimpleLayoutPass
from bqskit.passes.partitioning.cluster import ClusteringPartitioner
from bqskit.passes.partitioning.greedy import GreedyPartitioner
from bqskit.passes.partitioning.quick import QuickPartitioner
from bqskit.passes.partitioning.scan import ScanPartitioner
from bqskit.passes.partitioning.single import GroupSingleQuditGatePass
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.processing.substitute import SubstitutePass
from bqskit.passes.processing.window import WindowOptimizationPass
from bqskit.passes.rules.cnot2cz import CNOTToCZPass
from bqskit.passes.rules.zxzxz import ZXZXZDecomposition
from bqskit.passes.search.frontier import Frontier
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.middleout import MiddleOutLayerGenerator
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.passes.search.generators.single import SingleQuditLayerGenerator
from bqskit.passes.search.generators.stair import StairLayerGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.passes.search.heuristics.astar import AStarHeuristic
from bqskit.passes.search.heuristics.dijkstra import DijkstraHeuristic
from bqskit.passes.search.heuristics.greedy import GreedyHeuristic
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.qfast import QFASTDecompositionPass
from bqskit.passes.synthesis.qpredict import QPredictDecompositionPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.passes.util.compress import CompressPass
from bqskit.passes.util.conversion import BlockConversionPass
from bqskit.passes.util.converttou3 import ToU3Pass
from bqskit.passes.util.random import SetRandomSeedPass
from bqskit.passes.util.record import RecordStatsPass
from bqskit.passes.util.unfold import UnfoldPass
from bqskit.passes.util.update import UpdateDataPass

__all__ = [
    'DoWhileLoopPass',
    'ForEachBlockPass',
    'IfThenElsePass',
    'PassPredicate',
    'ChangePredicate',
    'GateCountPredicate',
    'NotPredicate',
    'WhileLoopPass',
    'ClusteringPartitioner',
    'GreedyPartitioner',
    'ScanPartitioner',
    'QuickPartitioner',
    'SynthesisPass',
    'LEAPSynthesisPass',
    'QSearchSynthesisPass',
    'QFASTDecompositionPass',
    'QPredictDecompositionPass',
    'CompressPass',
    'RecordStatsPass',
    'SetRandomSeedPass',
    'UnfoldPass',
    'UpdateDataPass',
    'ToU3Pass',
    'SimpleLayoutPass',
    'ScanningGateRemovalPass',
    'WindowOptimizationPass',
    'SimpleLayerGenerator',
    'AStarHeuristic',
    'GreedyHeuristic',
    'DijkstraHeuristic',
    'Frontier',
    'LayerGenerator',
    'HeuristicFunction',
    'SeedLayerGenerator',
    'BlockConversionPass',
    'StairLayerGenerator',
    'DoThenDecide',
    'SubstitutePass',
    'ParallelDo',
    'LoadCheckpointPass',
    'SaveCheckpointPass',
    'SaveIntermediatePass',
    'RestoreIntermediatePass',
    'GroupSingleQuditGatePass',
    'SingleQuditLayerGenerator',
    'MiddleOutLayerGenerator',
    'CNOTToCZPass',
    'ZXZXZDecomposition',
]



"""This module implements the DoThenDecide class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable
from typing import Sequence

from bqskit.compiler.basepass import BasePass
from bqskit.ir.circuit import Circuit
from bqskit.utils.typing import is_sequence

_logger = logging.getLogger(__name__)


class DoThenDecide(BasePass):
    """
    The DoThenDecide class.

    This is a control pass that executes a pass or sequence of passes and then
    conditionally accepts it or reverts to the original state.
    """

    def __init__(
        self,
        condition: Callable[[Circuit, Circuit], bool],
        pass_list: BasePass | Sequence[BasePass],
    ) -> None:
        """
        Construct a DoThenDecide.

        Args:
            condition (Callable[[Circuit, Circuit], bool]): The condition
                function that determines if the new circuit
                (second parameter) after running `pass_list` should replace
                the original circuit (first parameter). If the condition
                returns True, then replaced the original circuit with the
                new one.

            pass_list (BasePass | Sequence[BasePass]): The pass or passes
                to execute.

        Raises:
            ValueError: If a Sequence[BasePass] is given, but it is empty.
        """
        if not is_sequence(pass_list) and not isinstance(pass_list, BasePass):
            raise TypeError(
                'Expected Pass or sequence of Passes, got %s.'
                % type(pass_list),
            )

        if is_sequence(pass_list):
            truth_list = [isinstance(elem, BasePass) for elem in pass_list]
            if not all(truth_list):
                raise TypeError(
                    'Expected Pass or sequence of Passes, got %s.'
                    % type(pass_list[truth_list.index(False)]),
                )
            if len(pass_list) == 0:
                raise ValueError('Expected at least one pass.')

        self.condition = condition
        self.pass_list = pass_list if is_sequence(pass_list) else [pass_list]

    def run(self, circuit: Circuit, data: dict[str, Any] = {}) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # Perform Work
        old_circuit = circuit.copy()
        _logger.debug('Pass list executing...')
        for bqskit_pass in self.pass_list:
            bqskit_pass.run(circuit, data)

        if self.condition(old_circuit, circuit):
            _logger.debug('Accepted circuit.')

        else:
            circuit.become(old_circuit)
            _logger.debug('Rejected circuit.')
